import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import os
from .Point import Point, VirtualPoint, MarkerPoint


class Plane:
    def __init__(self, pt1=None, pt2=None, pt3=None):
        self.random_id = np.random.randint(0, 100000000)
        self.is_empty = True
        if pt1 is not None and pt2 is not None and pt3 is not None:
            self.set_by_pts(pt1, pt2, pt3)

    def set_by_pts(self, pt1, pt2, pt3):
        """
        rhr for vec: pt1->pt2, pt1->pt3
        try to make orthogonal vector positive for one axis
        """
        self.pt1 = pt1
        self.pt2 = pt2
        self.pt3 = pt3
        self.is_empty = False
        self.normal_vector = Point.orthogonal_vector(pt1, pt2, pt3, normalize=1)
        self.normal_vector_end = Point.translate_point(pt1, self.normal_vector, direction=1)

    def set_by_vector(self, pt1, vector, direction=1):
        """
        vector as virtual point
        """
        self.pt1 = pt1
        self.pt2 = None
        self.pt3 = None
        self.is_empty = False
        vector_xyz = vector.xyz
        normal_vector_xyz = vector_xyz / np.linalg.norm(vector_xyz, axis=0) * direction  # normalize vector
        self.normal_vector = Point.point_from_nparray(normal_vector_xyz)
        self.normal_vector_end = Point.translate_point(pt1, vector, direction=direction)

    def project_vector(self, vector, optimize=True):
        """
        project a vector onto the plane
        vector as xyz
        """
        plane_normal = self.normal_vector.xyz
        plane_normal = plane_normal / np.linalg.norm(plane_normal, axis=0)

        # vector = np.array([[1,1,1],[1,1,2]]).T
        # plane_normal = np.array([[0,1,0],[0,1,0]]).T
        # angle = Point.angle(vector, plane_normal)
        if optimize:  # optimized for memory
            diagonal_elements = np.einsum('ij,ij->i', vector.T, plane_normal.T)
            projection = vector - diagonal_elements[None, :] * plane_normal
        else:
            projection = vector - np.diagonal(np.dot(vector.T, plane_normal)) * plane_normal
        return projection

    def project_point(self, point):
        """
        project a point onto the plane
        """
        vector = point.xyz - self.pt1.xyz
        projection = self.project_vector(vector)  # todo: find a less entangled way in the future
        return Point.translate_point(self.pt1, projection)

    def above_or_below(self, point):
        """
        return 1 if point is above the plane, -1 if below
        """
        vector = point.xyz - self.pt1.xyz
        normal_vector = self.normal_vector.xyz
        return np.sign(np.sum(vector * normal_vector, axis=0))

    @staticmethod
    def angle_w_direction(plane1, plane2):
        '''
        return the angle between plane1 and plane2 in range of [-pi, pi]
        '''
        angle = Point.angle(plane1.normal_vector, plane2.normal_vector)
        angle_sign = Point.angle(plane1.normal_vector, plane2.normal_vector_end)
        angle = np.where(angle_sign > np.pi / 2, -angle, angle)
        return angle


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import os
from .Point import Point, VirtualPoint, MarkerPoint
from .Plane import Plane


class JointAngles:
    def __init__(self):
        self.random_id = np.random.randint(0, 100000000)
        self.is_empty = True
        self.zero_frame_or_angle = [None, None, None]
        self.zero_by_frame_not_angle = False
        self.ergo_name = {'flexion':'flexion', 'abduction':'abduction', 'rotation':'rotation'}

    def set_zero(self, frame_or_angle, by_frame=False):
        '''
        set to None if you don't want to zero the angles
        if by_frame is True, zero_frame_or_angle is a list of 3 frame numbers
        if by_frame is False, zero_frame_or_angle is a list of 3 angles in degrees
        '''
        if frame_or_angle is None:
            output = [None, None, None]
        elif type(frame_or_angle) == list:
            if len(frame_or_angle) != 3:
                raise ValueError('zero frame must be a list of length 3 or a single int')
            output = [np.radians(x) for x in frame_or_angle] if by_frame is False else frame_or_angle
        elif type(frame_or_angle) == int:
            frame_or_angle = np.radians(frame_or_angle) if by_frame is False else frame_or_angle  # convert to radian
            output = [frame_or_angle, frame_or_angle, frame_or_angle]
        self.zero_frame_or_angle = output
        self.zero_by_frame_not_angle = by_frame

    def get_flex_abd(self, coordinate_system, target_vector, plane_seq=['xy', 'xz'], flip_sign=[1, 1]):
        """
        get flexion and abduction angles of a vector in a coordinate system
        plane_seq: ['xy', None]
        """
        if len(plane_seq) != 2:
            raise ValueError('plane_seq must be a list of length 2, with flexion plane first and abduction plane second, fill None if not needed')
        xy_angle, xz_angle, yz_angle = coordinate_system.projection_angles(target_vector)
        output_angles = []
        for plane_id, plane_name in enumerate(plane_seq):
            if plane_name is not None:
                if plane_name == 'xy':
                    output_angle = xy_angle
                elif plane_name == 'xz' or plane_name == 'zx':
                    output_angle = xz_angle
                elif plane_name == 'yz':
                    output_angle = yz_angle
                else:
                    raise ValueError('plane_name must be one of "xy", "xz", "zx", "yz", or None')
                # output_angle = np.abs(output_angle)
                if self.zero_frame_or_angle[plane_id] is not None:
                    if self.zero_by_frame_not_angle:
                        zero_frame_id = self.zero_frame_or_angle[plane_id]
                        zero_angle = output_angle[zero_frame_id]
                    else:  # zero by angle directly
                        zero_angle = self.zero_frame_or_angle[plane_id]
                    output_angle = output_angle - zero_angle
                    # deal with output -pi pi range issue
                    output_angle = np.where(output_angle > np.pi, output_angle - 2 * np.pi, output_angle)
                    output_angle = np.where(output_angle < -np.pi, output_angle + 2 * np.pi, output_angle)
                output_angles.append(output_angle)

            else:
                output_angles.append(None)

        self.flexion = output_angles[0] * flip_sign[0]
        # self.flexion_info = {'plane': plane_seq[0], 'zero_angle': zero_angles[0], 'zero_frame': self.zero_frame[0], 'flip_sign': flip_sign[0]}
        self.abduction = output_angles[1] * flip_sign[1]
        # self.abduction_info = {'plane': plane_seq[1], 'zero_angle': zero_angles[1], 'zero_frame': self.zero_frame[1], 'flip_sign': flip_sign[1]}
        self.is_empty = False
        return output_angles

    def get_rot(self, pt1a, pt1b, pt2a, pt2b, flip_sign=1):
        '''
        get rotation angle between two vectors
        flip_sign: 1 or -1, if the rotation is in the opposite direction
        Example:

        '''
        pt1mid = Point.mid_point(pt1a, pt1b)
        pt2mid = Point.mid_point(pt2a, pt2b)
        plane1 = Plane(pt1a, pt1b, pt2mid)
        plane2 = Plane(pt2a, pt2b, pt1mid)
        rotation_angle = Point.angle(plane1.normal_vector.xyz, plane2.normal_vector.xyz)

        rotation_sign = plane2.above_or_below(pt1a)
        rotation_angle = rotation_angle * rotation_sign * flip_sign

        if self.zero_frame_or_angle[2] is not None:
            if self.zero_by_frame_not_angle:
                rotation_zero = rotation_angle[int(self.zero_frame_or_angle[2])]
            else:
                rotation_zero = self.zero_frame_or_angle[2]
            rotation_angle = rotation_angle - rotation_zero
        else:
            rotation_zero = None

        # make plot in -pi to pi range
        rotation_angle = np.where(rotation_angle > np.pi, rotation_angle - 2 * np.pi, rotation_angle)
        rotation_angle = np.where(rotation_angle < -np.pi, rotation_angle + 2 * np.pi, rotation_angle)

        self.rotation = rotation_angle
        self.rotation_info = {'plane': None, 'zero_angle': rotation_zero, 'zero_frame': self.zero_frame_or_angle[2]}
        self.is_empty = False
        return self.rotation

    def zero_by_idx(self, idx):
        """
        idx is 0, 1, 2, corresponding to flexion, abduction, rotation
        usage:
        RKNEE_angles.flexion = RKNEE_angles.zero_by_idx(0)
        """
        angle = self.flexion if idx == 0 else self.abduction if idx == 1 else self.rotation
        this_zero_frame = self.zero_frame_or_angle[idx]
        if this_zero_frame is not None:
            if self.zero_by_frame_not_angle:
                zero_angle = angle[this_zero_frame]
            else:
                zero_angle = this_zero_frame
            output_angle = angle - zero_angle
            output_angle = np.where(output_angle > np.pi, output_angle - 2 * np.pi, output_angle)
            output_angle = np.where(output_angle < -np.pi, output_angle + 2 * np.pi, output_angle)
        else:
            output_angle = angle

        return output_angle

    def output_np(self):
        output_list = []
        if self.flexion is not None:
            output_list.append(self.flexion)
        if self.abduction is not None:
            output_list.append(self.abduction)
        if self.rotation is not None:
            output_list.append(self.rotation)
        return np.array(output_list)

    def plot_angles(self, joint_name='', alpha=1, linewidth=1, linestyle='-', label=None, frame_range=None, colors=['r', 'g', 'b'], overlay=None):
        """
        plot angles
        overlay: [fig, ax] to overlay on an existing plot
        """
        if self.is_empty:
            raise ValueError('JointAngles is empty, please set angles first')
        if frame_range is None:
            if self.flexion is not None:
                frame_range = [0, len(self.flexion)]
            elif self.abduction is not None:
                frame_range = [0, len(self.abduction)]
            elif self.rotation is not None:
                frame_range = [0, len(self.rotation)]
            else:
                raise ValueError('all three angles are None, cannot plot')
        if overlay is None:
            fig, ax = plt.subplots(3, 1, sharex=True)
        else:
            fig, ax = overlay
        angle_names = ['Flexion', 'H-Abduction', 'Rotation']
        for angle_id, angle in enumerate([self.flexion, self.abduction, self.rotation]):
            # horizontal line at zero, pi, and -pi
            ax[angle_id].axhline(0, color='k', linestyle='--', alpha=0.5, linewidth=0.25)
            ax[angle_id].axhline(90, color='k', linestyle='dotted', alpha=0.5, linewidth=0.25)
            ax[angle_id].axhline(180, color='k', linestyle='--', alpha=0.5, linewidth=0.25)
            ax[angle_id].axhline(-90, color='k', linestyle='dotted', alpha=0.5, linewidth=0.25)
            ax[angle_id].axhline(-180, color='k', linestyle='--', alpha=0.5, linewidth=0.25)
            ax[angle_id].yaxis.set_ticks(np.arange(-180, 181, 90))
            ax[angle_id].set_ylabel(f'{angle_names[angle_id]}')
            ax[angle_id].set_xlim(frame_range[0], frame_range[1])  # set xlim
            ax[angle_id].margins(x=0)
            if angle is not None:
                ax[angle_id].plot(angle[0:frame_range[1]] / np.pi * 180, color=colors[angle_id], alpha=alpha, linewidth=linewidth, linestyle=linestyle, label=label)
            else:
                # plot diagonal line crossing through the chart
                ax[angle_id].plot([frame_range[0], frame_range[1]], [-180, 180], color='gray', linewidth=1)

        ax[0].set_title(f'{joint_name} (deg)')
        # plt.show()
        return fig, ax

    def plot_angles_by_frame(self, render_dir, joint_name='', alpha=1, linewidth=1, linestyle='-', label=None, frame_range=None, angle_names=['Flexion', 'H-Abduction', 'Rotation']):
        if self.is_empty:
            raise ValueError('JointAngles is empty, please set angles first')
        if frame_range is None:
            if self.flexion is not None:
                frame_range = [0, len(self.flexion)]
            elif self.abduction is not None:
                frame_range = [0, len(self.abduction)]
            elif self.rotation is not None:
                frame_range = [0, len(self.rotation)]
            else:
                raise ValueError('all three angles are None, cannot plot')
        print(f'Saving {joint_name} angle frames to {render_dir}')
        for frame_id in range(frame_range[0], frame_range[1]):
            fig, ax = plt.subplots(3, 1, sharex=True)

            colors = ['r', 'g', 'b']
            for angle_id, angle in enumerate([self.flexion, self.abduction, self.rotation]):
                print(f'frame {frame_id}/{frame_range[1]}', end='\r')
                # horizontal line at zero, pi, and -pi
                ax[angle_id].axhline(0, color='k', linestyle='--', alpha=0.5, linewidth=0.25)
                ax[angle_id].axhline(90, color='k', linestyle='dotted', alpha=0.5, linewidth=0.25)
                ax[angle_id].axhline(180, color='k', linestyle='--', alpha=0.5, linewidth=0.25)
                ax[angle_id].axhline(-90, color='k', linestyle='dotted', alpha=0.5, linewidth=0.25)
                ax[angle_id].axhline(-180, color='k', linestyle='--', alpha=0.5, linewidth=0.25)
                ax[angle_id].yaxis.set_ticks(np.arange(-180, 181, 90))
                ax[angle_id].set_xlim(frame_range[0], frame_range[1])  # set xlim
                ax[angle_id].axvline(frame_id, color='k', linestyle='--', alpha=0.5, linewidth=0.25)  # vertical line at current frame
                # a dot with value at current frame
                ax[angle_id].set_ylabel(f'{angle_names[angle_id]}')
                ax[angle_id].margins(x=0)
                if angle is not None:
                    ax[angle_id].plot(angle[0:frame_id + 1] / np.pi * 180, color=colors[angle_id], alpha=alpha, linewidth=linewidth, linestyle=linestyle, label=label)
                    ax[angle_id].plot(frame_id, angle[frame_id] / np.pi * 180, color=colors[angle_id], marker='o', markersize=5)  # a dot with value at current frame
                    ax[angle_id].text(frame_id, angle[frame_id] / np.pi * 180, f'{angle[frame_id] / np.pi * 180:.1f}', fontsize=12, horizontalalignment='left',
                                      verticalalignment='bottom')  # add text of current angle value
                else:
                    ax[angle_id].plot([frame_range[0], frame_range[1]], [-180, 180], color='gray', linewidth=1)  # plot diagonal line crossing through the chart

            ax[0].set_title(f'{joint_name} (deg)')

            ax[2].xaxis.set_major_locator(MaxNLocator(integer=True))  # set x ticks to integer only
            if not os.path.exists(render_dir):
                os.makedirs(render_dir)
            plt.savefig(os.path.join(render_dir, f'{joint_name}_{frame_id:06d}.png'))
            plt.close()


# todo: make them not senstive to 180 overflow errors.
def angle_diff(angle1, angle2, input_rad=True, output_rad=True):
    """
    smaller angle difference between two angles in circular space
    input -pi to pi range angles
    returns within -pi to pi range
    """
    if not input_rad:
        angle1 = np.deg2rad(angle1)
        angle2 = np.deg2rad(angle2)
    diff = angle1 - angle2
    diff = np.where(diff > np.pi, diff - 2 * np.pi, diff)
    diff = np.where(diff < -np.pi, diff + 2 * np.pi, diff)
    if not output_rad:
        diff = np.rad2deg(diff)
    return diff


def bland_altman_plot(data1, data2, title='', xlabel='Mean (\N{DEGREE SIGN})', ylabel='Difference (\N{DEGREE SIGN})', save_path=None,
                      plot_percentage=0.05, csfont={'fontname':'Times New Roman'}):
    """
    Bland-Altman plot
    data1, data2: two sets of measurements
    """
    data1 = np.rad2deg(data1)
    data2 = np.rad2deg(data2)
    mean = np.nanmean([data1, data2], axis=0)
    diff = angle_diff(data1, data2, input_rad=False, output_rad=False)
    md = np.nanmean(diff)
    sd = np.nanstd(diff, axis=0)
    # only plot 10% of the data
    idx = np.random.choice(range(len(mean)), int(len(mean) * plot_percentage), replace=False)
    plt.scatter(mean[idx], diff[idx], alpha=0.2, s=1)
    plt.axhline(md, color='red', linestyle='--')
    plt.axhline(md + 1.96 * sd, color='red', linestyle='--')
    plt.axhline(md - 1.96 * sd, color='red', linestyle='--')
    text_height = 1
    text_right_pos = 0.98 * (plt.gca().get_xlim()[1] - plt.gca().get_xlim()[0]) + plt.gca().get_xlim()[0]
    # text above the line on the right hand side "+1.9, weight='bold'6 SD, size
    t1 = plt.text(text_right_pos, md + 1.96 * sd + text_height, f'+1.96 SD: {md+1.96*sd:.1f}\N{DEGREE SIGN}', horizontalalignment='right', verticalalignment='bottom',  fontsize=19, weight='bold', **csfont)
    t2 = plt.text(text_right_pos, md - 1.96 * sd - text_height, f'-1.96 SD: {md-1.96*sd:.1f}\N{DEGREE SIGN}', horizontalalignment='right', verticalalignment='top',     fontsize=19, weight='bold', **csfont)
    t3 = plt.text(text_right_pos, md , f'Mean: {md:.1f}\N{DEGREE SIGN}',                                      horizontalalignment='right', verticalalignment='bottom',  fontsize=19, weight='bold', **csfont)
    for t in [t1, t2, t3]:
        t.set_bbox(dict(facecolor='white', alpha=0.5, edgecolor='white', linewidth=0))
    plt.xticks(fontsize=12)
    plt.yticks(fontsize=12)
    plt.title(title,    fontsize=28, weight='bold', **csfont)
    # plt.xlabel(xlabel,  fontsize=22, **csfont)
    # plt.ylabel(ylabel,  fontsize=22, **csfont)
    plt.ylim([-35, 35])
    # plt.text(0.5, 0.95, f'Mean diff: {md:.2f}, std {sd:.2f}', horizontalalignment='center', verticalalignment='center', transform=plt.gca().transAxes)
    if save_path is not None:
        plt.savefig(save_path)
    else:
        plt.show()
    plt.close()
    return md, sd


def mean_absolute_error(data1, data2):
    """
    data1, data2: two sets of measurements
    """
    diff = angle_diff(data1, data2, input_rad=True, output_rad=False)
    return np.nanmean(np.abs(diff))


def root_mean_squared_error(data1, data2):
    """
    data1, data2: two sets of measurements
    """
    diff = angle_diff(data1, data2, input_rad=True, output_rad=False)
    return np.nanmean(diff ** 2)**0.5


# def bland_altman_plot_batch(ja1, ja2, angle_names):
#     """
#     ja1, ja2: two sets of JointAngle objects
#     set ja2 to gt
#     """
#     for angle_index, this_angle_name in enumerate(angle_names):
#
#         # md, sd = bland_altman_plot(ja1[this_angle_name].flexion, ja2[this_angle_name].flexion, title=this_angle_name) #, save_path=f'frames/MB_angles/{this_angle_name}_bland_altman.png')
#         # print(f'Visualizing - {this_angle_name}: md: {md:.2f}, sd: {sd:.2f}')
#         MSE, sd = mean_squared_error(ja1[this_angle_name].flexion, ja2[this_angle_name].flexion)
#         print(f'MSE - {this_angle_name}: mean: {MSE:.2f}, sd: {sd:.2f}')
#
#         MAE = mean_absolute_error(ja1[this_angle_name].flexion, ja2[this_angle_name].flexion)
#         print(f'MAE - {this_angle_name}: {MAE:.2f}')

# bland_altman_plot_batch(estimate_ergo_angles, GT_ergo_angles, GT_skeleton.angle_names)
#
# ja1 = GT_ergo_angles
# ja2 = estimate_ergo_angles
#
# this_angle_name = 'left_wrist'
# for i, angle in enumerate(estimate_ergo_angles[this_angle_name].flexion):
#     if np.isnan(angle):
#         print(i)


import xml.etree.ElementTree as ET
import os
import numpy as np
import cv2
import pandas as pd
import json
import warnings


class Camera():
    def __int__(self):
        warnings.warn("All quaternions are in the form [x, y, z, w], double check your input")

    @staticmethod
    def string_to_float(string, delimiter=' ', suppress_warning=False):
        """
        Convert a string of floats separated by a delimiter to a list of floats.
        :param string: 
        :param delimiter: 
        :param suppress_warning: 
        :return: 
        """
        output = []
        for x in string.split(delimiter):
            try:
                output.append(float(x))
            except ValueError:
                if not suppress_warning:
                    print(f'Warning: \'{x}\' in string \'{string} \'is not a float.')
        if len(output) == 1:
            return output[0]
        else:
            return np.array(output)

    @staticmethod
    def rot3x3_from_quaternion(quaternion):
        x, y, z, w = quaternion
        rot_matrix = np.array([[1 - 2 * y ** 2 - 2 * z ** 2, 2 * x * y - 2 * z * w, 2 * x * z + 2 * y * w],
                               [2 * x * y + 2 * z * w, 1 - 2 * x ** 2 - 2 * z ** 2, 2 * y * z - 2 * x * w],
                               [2 * x * z - 2 * y * w, 2 * y * z + 2 * x * w, 1 - 2 * x ** 2 - 2 * y ** 2]])
        return rot_matrix

    @staticmethod
    def quaternion_from_rot3x3(rot3x3):  # todo: cover with test, wrong current, or imcommutable at least
        """
        Convert a 3x3 rotation matrix to a quaternion.

        :param rot3x3: A 3x3 rotation matrix.
        :return: A quaternion [w, x, y, z].  # tido: also not following the x, y, z, w convention in this package
        """
        raise NotImplementedError('Not working correctly, do not use')

        # Ensure the input is a numpy array for matrix operations.
        R = np.array(rot3x3)

        # Allocate space for the quaternion.
        q = np.zeros(4)

        # Compute the trace of the matrix.
        tr = np.trace(R)

        if tr > 0:
            S = np.sqrt(tr + 1.0) * 2  # S=4*qw 
            q[0] = 0.25 * S
            q[1] = (R[2, 1] - R[1, 2]) / S
            q[2] = (R[0, 2] - R[2, 0]) / S
            q[3] = (R[1, 0] - R[0, 1]) / S
        elif (R[0, 0] > R[1, 1]) and (R[0, 0] > R[2, 2]):
            S = np.sqrt(1.0 + R[0, 0] - R[1, 1] - R[2, 2]) * 2  # S=4*qx
            q[0] = (R[2, 1] - R[1, 2]) / S
            q[1] = 0.25 * S
            q[2] = (R[0, 1] + R[1, 0]) / S
            q[3] = (R[0, 2] + R[2, 0]) / S
        elif R[1, 1] > R[2, 2]:
            S = np.sqrt(1.0 + R[1, 1] - R[0, 0] - R[2, 2]) * 2  # S=4*qy
            q[0] = (R[0, 2] - R[2, 0]) / S
            q[1] = (R[0, 1] + R[1, 0]) / S
            q[2] = 0.25 * S
            q[3] = (R[1, 2] + R[2, 1]) / S
        else:
            S = np.sqrt(1.0 + R[2, 2] - R[0, 0] - R[1, 1]) * 2  # S=4*qz
            q[0] = (R[1, 0] - R[0, 1]) / S
            q[1] = (R[0, 2] + R[2, 0]) / S
            q[2] = (R[1, 2] + R[2, 1]) / S
            q[3] = 0.25 * S

        return q



    @staticmethod
    def rot3x4_from_rot3x3_and_position(rot3x3, position):
        rot3x4 = np.hstack([rot3x3, -rot3x3.dot(position.reshape(-1, 1))])
        return rot3x4


    @staticmethod
    def quaternion_from_axis_angle(axis, angle=None):
        """
        Axis angle to quaternion, for SMPL mostly
        :param axis: (3,)
        :param angle: None or float
        :return:
        """
        axis_norm = np.linalg.norm(axis)
        angle = axis_norm if angle is None else angle
        axis = axis / axis_norm if axis_norm > 1e-8 else np.array([0, 0, 0])  # todo: sin(x/2)/x ~= 1/2 - (x**2)/48 for small angle x < 1e6
        w = np.cos(angle / 2.0)
        xyz = axis * np.sin(angle / 2.0)
        return np.array([xyz[0], xyz[1], xyz[2], w])

    @staticmethod
    def axis_angle_from_quaternion(quaternion, one_output=True):
        """
        Quaternion to axis angle, for SMPL mostly
        :param quaternion: (4,) x, y, z, w
        :param one_output: True: put angle as the norm of the axis vector
        :return:
        """
        x, y, z, w = quaternion
        angle = 2 * np.arccos(w)
        sin_theta_over_two = np.sqrt(1 - w * w)
        # To avoid division by zero, check if sin(theta/2) is close to zero
        axis = np.array([x, y, z]) / sin_theta_over_two if sin_theta_over_two > 1e-8 else np.array([1, 0, 0]) # If the angle is close to 0, direction does not matter  # todo: small angle approximation
        return axis * angle if one_output else [axis, angle]

    @staticmethod
    def quaternion_multiply(q1, q2):
        """
        Multiply two quaternions.

        Parameters:
        - q1: A tuple or list representing the first quaternion (x, y, z, w).
        - q2: A tuple or list representing the second quaternion (x, y, z, w).

        Returns:
        - A tuple representing the product of the two quaternions.
        """
        x1, y1, z1, w1 = q1
        x2, y2, z2, w2 = q2

        # Calculate the product
        w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
        x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
        y = w1 * y2 - x1 * z2 + y1 * w2 + z1 * x2
        z = w1 * z2 + x1 * y2 - y1 * x2 + z1 * w2

        return np.array([x, y, z, w])

    @staticmethod
    def rotate_axis_angle_by_axis_angle(axis_angle, axis_angle_delta):
        """
        Rotate an axis-angle vector by another axis-angle vector, for SMPL mostly
        :param axis_angle: the rotation vector, magnitude is the angle turned anticlockwise (RHR) in radians around the vector's direction
        :param axis_angle_delta:
        :return:
        """
        q1 = Camera.quaternion_from_axis_angle(axis_angle)
        q2 = Camera.quaternion_from_axis_angle(axis_angle_delta)
        q_out = Camera.quaternion_multiply(q2, q1)
        return Camera.axis_angle_from_quaternion(q_out, one_output=True)

    def get_camera_intrinsic_matrix(self):
        # todo: check if the skew is correct, currently all cameras have skew 0.0
        K = np.array([[self.FOCAL_LENGTH, self.SKEW, self.PRINCIPAL_POINT[0]],
                      [0, self.FOCAL_LENGTH / self.PIXEL_ASPECT_RATIO, self.PRINCIPAL_POINT[1]],
                      [0, 0, 1]])
        return K

    @staticmethod
    def projection_matrix_from_intrinsic_matrix_and_rot3x4(K, rot3x4):
        P = K.dot(rot3x4)
        return P

    def undistort(self, points_2d):
        raise NotImplementedError

    def distort(self, points_2d):
        raise NotImplementedError

    def get_projection_matrix(self):
        if not hasattr(self, 'rot3x3'):
            self.rot3x3 = self.rot3x3_from_quaternion(self.ORIENTATION)
        self.rot3x4 = self.rot3x4_from_rot3x3_and_position(self.rot3x3, self.POSITION)
        self.intrinsic_matrix = self.get_camera_intrinsic_matrix()
        self.projection_matrix = self.projection_matrix_from_intrinsic_matrix_and_rot3x4(self.intrinsic_matrix, self.rot3x4)
        return self.projection_matrix

    def get_camera_pitch(self):
        """
        For pitch correction, assume small roll
        Compare gravity vector in camera coordinate system with camera y axis
        only work for one frame, use def project_w_depth in the future when multiple frames are needed
        """
        world_z_point = np.array([0, 0, 1])
        world_origin = np.array([0, 0, 0])

        self.get_projection_matrix()
        camera_z_point = np.dot(self.rot3x4, np.hstack([world_z_point, 1]))
        camera_origin = np.dot(self.rot3x4, np.hstack([world_origin, 1]))

        camera_gravity_vector = camera_origin - camera_z_point  # +z-->origin
        camera_y_axis = np.array([0, 1, 0])
        pitch_correction_angle = np.arccos(np.dot(camera_gravity_vector, camera_y_axis) / (np.linalg.norm(camera_gravity_vector) * np.linalg.norm(camera_y_axis)))
        return pitch_correction_angle

    def project(self, points_3d):
        ''' extrinsic and intrinsic projection
        :param points_3d: 3d points in world coordinate
        :return: 2d points in camera coordinate, in pixel
        '''
        self.get_projection_matrix()
        points_3d = np.array(points_3d)
        points_2d = self.projection_matrix.dot(np.vstack([points_3d.T, np.ones([1, points_3d.shape[0]])]))
        points_2d = points_2d[0:2, :] / points_2d[2, :]
        points_2d = points_2d.T
        return points_2d

    def project_w_depth(self, points_3d):
        ''' world to camera coordinate system, extrinsic projection
        :param points_3d: 3d points in world coordinate
        :return: 3d points in camera coordinate, with depth, in mm/m
        '''
        self.get_projection_matrix()
        points_3d = np.array(points_3d)
        points_3d_cam = self.rot3x4.dot(np.vstack([points_3d.T, np.ones([1, points_3d.shape[0]])]))
        # translation_vector = self.POSITION.reshape([3, 1])
        # rotM = self.rot3x3
        # points_3d_cam = (rotM.dot(points_3d.T) + translation_vector).T
        return points_3d_cam.T

    def _weak_project(self, pose3d):
        """ From LCN, intrinsic projection
        """
        fx, fy = self.FOCAL_LENGTH, self.FOCAL_LENGTH / self.PIXEL_ASPECT_RATIO
        cx, cy = self.PRINCIPAL_POINT
        pose2d = pose3d[:, :2] / pose3d[:, 2:3]
        pose2d[:, 0] *= fx
        pose2d[:, 1] *= fy
        pose2d[:, 0] += cx
        pose2d[:, 1] += cy
        return pose2d

    def draw_2d(self, image, points_2d, color=(0, 0, 255)):
        for point_idx, point_2d in enumerate(points_2d):
            cv2.circle(image, (int(point_2d[0]), int(point_2d[1])), 3, color, -1)
            cv2.putText(image, str(point_idx), (int(point_2d[0]), int(point_2d[1])), cv2.FONT_HERSHEY_SIMPLEX, 0.5, color, 1)
        return image



from .Camera import *


def batch_load_from_xcp(xcp_filename):
    # Read the xcp file as xml
    with open(xcp_filename, 'r') as f:
        xml_string = f.read()
    root = ET.fromstring(xml_string)
    cameras = []
    for child in root:
        if child.attrib['DISPLAY_TYPE'] == 'VideoInputDevice:Blackfly S BFS-U3-23S3C':
            camera = FLIR_Camera()
            camera.load_vicon_xcp(child)
            cameras.append(camera)
    return cameras


def load_csv(csv_filename):
    df = pd.read_csv(csv_filename, skiprows=[0, 1, 2,3, 4], header=None)
    header_df = pd.read_csv(csv_filename, skiprows=[0, 1], nrows=1, header=None)
    # remove nan in the header
    # iterate over header_df
    header = ['Frame', 'Sub Frame']
    axis = ['X', 'Y', 'Z']
    joint_names = []
    for x in header_df.iloc[0]:
        # print(x)
        if type(x) == str:
            joint_name = x.split(':')[-1]
            joint_names.append(joint_name)
            for ax in axis:
                header.append(f'{joint_name}-{ax}')
    # add header to df
    df.columns = header
    return df, joint_names


class FLIR_Camera(Camera):
    def __init__(self):
        super().__init__()

    def load_vicon_xcp(self, xcp_element):
        # Read the xcp file as xml
        self.DEVICEID = xcp_element.attrib['DEVICEID']
        self.SENSOR_SIZE = self.string_to_float(xcp_element.attrib['SENSOR_SIZE'])
        self.DISPLAY_TYPE = xcp_element.attrib['DISPLAY_TYPE']
        self.USERID = xcp_element.attrib['USERID']
        self.PIXEL_ASPECT_RATIO = float(xcp_element.attrib['PIXEL_ASPECT_RATIO'])
        for grandchild in xcp_element:
            if grandchild.tag == 'ThresholdGrid':
                pass
            elif grandchild.tag == 'Calibration':
                pass
            elif grandchild.tag == 'IntrinsicsCalibration':
                pass
            elif grandchild.tag == 'Capture':
                self.CAPTURE_START_TIME = grandchild.attrib['START_TIME']
                self.CAPTURE_END_TIME = grandchild.attrib['END_TIME']
            elif grandchild.tag == 'KeyFrames':
                for grandchild2 in grandchild:
                    self.FOCAL_LENGTH = self.string_to_float(grandchild2.attrib['FOCAL_LENGTH'])
                    self.IMAGE_ERROR = self.string_to_float(grandchild2.attrib['IMAGE_ERROR'])
                    self.ORIENTATION = self.string_to_float(grandchild2.attrib['ORIENTATION'])
                    self.POSITION = self.string_to_float(grandchild2.attrib['POSITION'])/1000
                    self.PRINCIPAL_POINT = self.string_to_float(grandchild2.attrib['PRINCIPAL_POINT'])
                    self.VICON_RADIAL2 = self.string_to_float(grandchild2.attrib['VICON_RADIAL2'], suppress_warning=True)
                    self.DISTORTION_CENTER = self.VICON_RADIAL2[0:2]
                    self.DISTORTION_SCALE = np.hstack([np.array([1]), self.VICON_RADIAL2[2:]])
                    self.WORLD_ERROR = self.string_to_float(grandchild2.attrib['WORLD_ERROR'])
                    self.frame = self.string_to_float(grandchild2.attrib['FRAME'])
                    break
                if len(grandchild) > 1:
                    # raise warning
                    print('Warning: more than one keyframe found, is this a moving camera? Only first keyframe will be used.')
                    # todo: accommodate moving camera in the future
        self.SKEW = 0  # not exist in xcp file, assuming no skew

    def get_matrix(self):
        pass

    def undistort(self, points_2d):
        # undistort radial distortion, formula provided by Vicon
        points_2d = np.array(points_2d)
        dp = (points_2d - self.DISTORTION_CENTER) * (np.array([1, self.PIXEL_ASPECT_RATIO]))
        radius = np.linalg.norm(dp, axis=1)
        scale = np.vstack([np.ones_like(radius), radius ** 2, radius ** 4, radius ** 6]).T.dot(self.DISTORTION_SCALE).reshape(-1, 1)
        point_2d_corrected = (scale*dp + self.DISTORTION_CENTER)*(np.array([1, 1/self.PIXEL_ASPECT_RATIO]))  # assuming distortion center is the same as principal point
        return point_2d_corrected

    def distort(self, points_2d):
        # approximate distort radial distortion, scale calculation should use distorted points, but here we use undistorted points
        points_2d = np.array(points_2d)
        dp = points_2d * (np.array([1, self.PIXEL_ASPECT_RATIO])) - self.DISTORTION_CENTER
        radius = np.linalg.norm(dp, axis=1)
        scale = np.vstack([np.ones_like(radius), radius ** 2, radius ** 4, radius ** 6]).T.dot(self.DISTORTION_SCALE).reshape(-1, 1)
        point_2d_distorted = dp/scale*(np.array([1, 1/self.PIXEL_ASPECT_RATIO])) + self.DISTORTION_CENTER # assuming distortion center is the same as principal point
        return point_2d_distorted


    @staticmethod
    def plugingait_to_H36M_converter(points_2d, frame_no, joint_names):
#         [{"image_id": "0.jpg",
#         "category_id": 1,
#         "keypoints": [371.8258361816406, 184.70167541503906, 0.9479594826698303, 383.28143310546875, 173.24606323242188, 0.9629446268081665, 366.0980224609375, 173.24606323242188, 0.9532919526100159, 411.92047119140625, 178.97386169433594, 0.9926742315292358, 360.3702087402344, 178.97386169433594, 0.8256455659866333, 440.55950927734375, 241.979736328125, 0.9333735704421997, 348.91461181640625, 241.979736328125, 0.9294862151145935, 474.92633056640625, 322.1690368652344, 0.9467807412147522, 343.1867980957031, 327.8968200683594, 0.9887729287147522, 434.8316955566406, 293.5299987792969, 0.9161380529403687, 331.7311706542969, 304.985595703125, 0.931166410446167, 423.3760681152344, 385.1748962402344, 0.9257044196128845, 371.8258361816406, 385.1748962402344, 0.8760479688644409, 434.8316955566406, 528.3700561523438, 0.8742569088935852, 348.91461181640625, 528.3700561523438, 0.9642834067344666, 469.1985168457031, 631.4705810546875, 0.901423990726471, 360.3702087402344, 637.1983642578125, 0.9199998378753662, 383.28143310546875, 144.60702514648438, 0.9656839966773987, 389.0092468261719, 219.06851196289062, 0.9374449849128723, 394.737060546875, 379.44708251953125, 0.9205844402313232, 452.0151062011719, 683.0208129882812, 0.9289825558662415, 348.91461181640625, 683.0208129882812, 0.9448415040969849, 469.1985168457031, 683.0208129882812, 0.9240862727165222, 337.458984375, 677.2930297851562, 0.9301682114601135, 474.92633056640625, 642.9262084960938, 0.8143528699874878, 371.8258361816406, 648.6539916992188, 0.8643205165863037],
#         "score": 3.155129909515381,
#         "box": [304.7809753417969, 129.14193725585938, 185.63992309570312, 586.5273742675781],
#         "idx": [0.0]},]
        # 17 joints of Human3.6M:
        # 'root', 'RHip', 'RKnee', 'RAnkle', 'LHip', 'LKnee', 'LAnkle', 'torso', 'neck', 'nose', 'head', 'LShoulder', 'LElbow', 'LWrist', 'RShoulder', 'RElbow', 'RWrist'
        # 17 joints of Plugingait model output:
        # BHEC, CentreOfMass, CentreOfMassFloor, FHEC, HDCC, HECC, HTPO, HeadCOM, LAJC, LEJC, LEYE, LHEC, LHJC, LKJC, LMFO, LPKO, LSJC, LWJC, LeftFemurCOM, LeftFootCOM, LeftHandCOM, LeftHumerusCOM, LeftRadiusCOM, LeftTibiaCOM, NKTP, PelvisCOM, RAJC, REJC, REYE, RHEC, RHJC, RKJC, RMFO, RPKO, RSJC, RWJC, RightFemurCOM, RightFootCOM, RightHandCOM, RightHumerusCOM, RightRadiusCOM, RightTibiaCOM, ThoraxCOM
        # w. index: 0. BHEC, 1. CentreOfMass, 2. CentreOfMassFloor, 3. FHEC, 4. HDCC, 5. HECC, 6. HTPO, 7. HeadCOM, 8. LAJC, 9. LEJC, 10. LEYE, 11. LHEC, 12. LHJC, 13. LKJC, 14. LMFO, 15. LPKO, 16. LSJC, 17. LWJC, 18. LeftFemurCOM, 19. LeftFootCOM, 20. LeftHandCOM, 21. LeftHumerusCOM, 22. LeftRadiusCOM, 23. LeftTibiaCOM, 24. NKTP, 25. PelvisCOM, 26. RAJC, 27. REJC, 28. REYE, 29. RHEC, 30. RHJC, 31. RKJC, 32. RMFO, 33. RPKO, 34. RSJC, 35. RWJC, 36. RightFemurCOM, 37. RightFootCOM, 38. RightHandCOM, 39. RightHumerusCOM, 40. RightRadiusCOM, 41. RightTibiaCOM, 42. ThoraxCOM
        H36M_frame = {}
        H36M_frame['image_id'] = f'{frame_no}.jpg'
        H36M_frame['category_id'] = 1
        H36M_frame['keypoints'] = []
        H36M_frame['score'] = 3.0
        H36M_frame['box'] = []
        H36M_frame['idx'] = [0.0]
        # selected plugingait joints in sequence:
        # PelvisCOM, RHJC, RKJC, RAJC, LHJC, LKJC, LACJ, 0.5*(HeadCOM+ThoraxCOM), ThoraxCOM, 0.5*(LEYE+REYE), HeadCOM, LSJC, LEJC, LWJC, RSJC, REJC, RWJC
        plugingait_joints = ['PelvisCOM', 'RHJC', 'RKJC', 'RAJC', 'LHJC', 'LKJC', 'LAJC', 'temp', 'ThoraxCOM', 'temp', 'HeadCOM', 'LSJC', 'LEJC', 'LWJC', 'RSJC', 'REJC', 'RWJC']
        keypoints = np.ones((17, 3))
        for i, joint_name in enumerate(plugingait_joints):
            if joint_name == 'temp':
                continue
            keypoints[i, :2] = points_2d[joint_names.index(joint_name)]
        keypoints[7, :2]    = 0.5*(points_2d[joint_names.index('HeadCOM')] + points_2d[joint_names.index('ThoraxCOM')])
        keypoints[9, :2]    = 0.5*(points_2d[joint_names.index('LEYE')] + points_2d[joint_names.index('REYE')])
        H36M_frame['keypoints'] = keypoints.reshape(-1).tolist()
        # find a bbox containing all keypoints
        H36M_frame['box'] = [np.min(keypoints[:, 0]), np.min(keypoints[:, 1]), np.max(keypoints[:, 0])-np.min(keypoints[:, 0]), np.max(keypoints[:, 1])-np.min(keypoints[:, 1])]
        return H36M_frame


if __name__ == '__main__':
    example_case = 1
    if example_case == 1:  # visualize rgb with overlay 2d skeleton
        base_dir_name = r'C:\Users\Public\Documents\Vicon\data\VEHS_ske\Round3\Leyang Wen\Subject 3-1'
        activity_no = 1
        frames = np.linspace(0, 6000, int(6000/5+1), dtype=int)
        time = '20230427162751'
        activity_name = f'Activity {activity_no:02d}'
        basename = os.path.join(base_dir_name, activity_name)
        output_dir = 'output/frames'
        c3d_filename = basename + '.c3d'
        xcp_filename = basename + '.xcp'
        csv_filename = basename + '.trajectory.csv'
        df, joint_names = load_csv(csv_filename)

        cameras = batch_load_from_xcp(xcp_filename)
        camera = cameras[1]
        for cam_idx, camera in enumerate(cameras):
            if cam_idx == 2:
                continue
            video_filename = f'{basename}.{camera.DEVICEID}.{time}.avi'
            # get first frame of video
            for frame_no in frames:
                print(f'Processing frame {frame_no}/{frames[-1]} of {activity_name}.{camera.DEVICEID}.{time}.avi', end='\r')
                points_3d = df.iloc[frame_no, 2:].values.reshape(-1, 3) / 1000
                cap = cv2.VideoCapture(video_filename)
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_no-1)
                ret, frame = cap.read()
                points_2d = camera.project(points_3d)
                # frame = camera.draw_2d(frame, points_2d)
                points_2d = camera.distort(points_2d)
                # points_2d = camera.undistort(points_2d)
                frame = camera.draw_2d(frame, points_2d, color=(0, 255, 0))
                # cv2.imshow('frame', frame)
                # cv2.waitKey(0)
                dir_name = os.path.join(output_dir, camera.DEVICEID)
                if not os.path.exists(dir_name):
                    os.makedirs(dir_name)
                output_file = os.path.join(dir_name, f'{activity_name}.{camera.DEVICEID}.{time}.frame{frame_no:06d}.jpg')
                cv2.imwrite(output_file, frame)

    if example_case == 2:  # output H36M json file for motionBert
        base_dir_name = r'C:\Users\Public\Documents\Vicon\vicon_coding_projects\MotionBERT\experiment\Vicon_Gunwoo_Test_movement02'
        activity_no = 2
        time = '20230224152348'  #'20230224152907'
        activity_name = f'Gunwoo movement {activity_no:02d}'
        basename = os.path.join(base_dir_name, activity_name)
        output_dir = f'{base_dir_name}'
        sep_frames = False
        xcp_filename = basename + '.xcp'
        csv_filename = basename + '.joint_loc.csv'
        df, joint_names = load_csv(csv_filename)
        fps = 50
        start_frame = 0
        end_frame = 7497
        fps_ratio = 100/fps
        frames = np.linspace(start_frame/fps_ratio, end_frame/fps_ratio, int((end_frame-start_frame)/fps_ratio), dtype=int)
        cameras = batch_load_from_xcp(xcp_filename)
        camera = cameras[1]
        for cam_idx, camera in enumerate(cameras):
            print(f'Processing camera {camera.DEVICEID}')
            if camera.DEVICEID == '66920731' or camera.DEVICEID == '66920758':
                pass
            else:
                continue
            json_list = []
            video_filename = f'{basename}.{camera.DEVICEID}.{time}.avi'
            for frame_idx, frame_no in enumerate(frames):
                frame_idx = int(frame_idx * fps_ratio)
                print(f'Processing frame {frame_no}/{frames[-1]} of {activity_name}.{camera.DEVICEID}.{time}.avi',
                      end='\r')
                points_3d = df.iloc[frame_idx, 2:].values.reshape(-1, 3) / 1000
                points_2d = camera.project(points_3d)
                points_2d = camera.distort(points_2d)
                json_line = camera.plugingait_to_H36M_converter(points_2d, frame_idx, joint_names)
                json_list.append(json_line)
                camera_points_2d = np.array(json_line['keypoints']).reshape(-1, 3)[:, :2]
                if sep_frames:
                    cap = cv2.VideoCapture(video_filename)
                    cap.set(cv2.CAP_PROP_POS_FRAMES, frame_no - 1)
                    ret, frame = cap.read()
                    if True:
                        frame = camera.draw_2d(frame, camera_points_2d, color=(0, 255, 0))
                        dir_name = os.path.join(output_dir, f'{camera.DEVICEID}_wJoints')
                    else:
                        dir_name = os.path.join(output_dir, camera.DEVICEID)
                    if not os.path.exists(dir_name):
                        os.makedirs(dir_name)
                    output_file = os.path.join(dir_name,
                                               f'{activity_name}.{camera.DEVICEID}.{time}.frame{int(frame_idx/2):06d}.jpg')
                    cv2.imwrite(output_file, frame)

            json_filename = os.path.join(output_dir, f'{activity_name}.{camera.DEVICEID}.json')
            with open(json_filename, 'w') as f:
                json.dump(json_list, f)




from setuptools import setup, find_packages

setup(
    name='ergo3d',  # Name of your package
    version='0.1.11',  # Version number
    description='A Python package for 3D ergonomic calculations.',  # Short description of your package
    url='https://github.com/LeyangWen/ergo3d',  # URL for your package's homepage
    author='Leyang Wen',  # Your name
    author_email='wenleyan@umich.edu',  # Your email
    license='MIT',  # License type for your package
    packages=find_packages(),  # Automatically find all packages and subpackages
    install_requires=[  # List of dependencies
        'numpy',
        'matplotlib',
        'opencv-python',
        'pandas'
    ],
    classifiers=[  # Classifiers help users find your project by categorizing it
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3.6',  # Minimum version of Python your package requires
)


# import ast
# import os
#
# def get_imports(filepath):
#     with open(filepath, 'r') as file:
#         root = ast.parse(file.read(), filepath)
#
#     for node in ast.iter_child_nodes(root):
#         if isinstance(node, ast.Import):
#             for alias in node.names:
#                 yield alias.name
#         elif isinstance(node, ast.ImportFrom):
#             yield node.module
#
# def get_all_imports_in_project(project_path):
#     imports = set()
#     for dirpath, dirnames, filenames in os.walk(project_path):
#         for filename in filenames:
#             if filename.endswith('.py'):
#                 filepath = os.path.join(dirpath, filename)
#                 for module in get_imports(filepath):
#                     imports.add(module)
#     return imports
#
# project_path = '.'  # path to your project
# all_imports = get_all_imports_in_project(project_path)
# print('\n'.join(sorted(all_imports)))



from Camera import *
# from .Camera import *
import cv2
import pickle
import os
import random

def batch_load_from_checkerboard(load_dir, file_extension='_back.pkl', cameras=[]):
    """
    Choose extension from ['_back.pkl', '_right.pkl', '_left.pkl'], append to cameras list
    """
    for root, dirs, files in os.walk(load_dir):
        files = [f for f in files if not f[0] == '.']  # skip hidden files
        for file in files:
            if not file.endswith(file_extension):
                continue
            print(f"Loading {file}...", end=' ::::::: ')
            camera = Checkerboard_Camera()
            camera.load(os.path.join(root, file))
            cameras.append(camera)
    return cameras


class Checkerboard_Camera(Camera):
    def __init__(self):
        super().__init__()
        self.criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)

    def copy(self):
        # make a new item and copy all attributes
        new_item = Checkerboard_Camera()
        new_item.__dict__ = self.__dict__.copy()
        return new_item

    def set_checkerboard_params(self, square_size, width=9, height=6):
        """
        Set the checkerboard parameters.
        :param square_size: size of the square in meters
        :param width: number of squares in the width
        :param height: number of squares in the height
        """
        self.square_size = square_size
        self.cb_width = width
        self.cb_height = height

    def set_camera_params(self, DEVICEID, VIEW_ROTATION=0):
        self.DEVICEID = DEVICEID
        self.VIEW_ROTATION = VIEW_ROTATION


    def intrinsic_calibrate(self, calibrate_img_dir, image_extension, verbose=True):
        """ Apply camera calibration operation for images in the given directory path. """
        # prepare object points, like (0,0,0), (1,0,0), (2,0,0) ....,(8,6,0)
        objp = np.zeros((self.cb_height * self.cb_width, 3), np.float32)
        objp[:, :2] = np.mgrid[0:self.cb_width, 0:self.cb_height].T.reshape(-1, 2)
        objp = objp * self.square_size  # Create real world coords. Use your metric.

        # Arrays to store object points and image points from all the images.
        objpoints = []  # 3d point in real world space
        imgpoints = []  # 2d points in image plane.


        for root, dirs, files in os.walk(calibrate_img_dir):
            files = [f for f in files if not f[0] == '.']  # skip hidden files
            dirs[:] = [d for d in dirs if not d[0] == '.']  # skip hidden folders
            dirs.sort()  # Sort directories in-place --> important, will change walk sequence
            files.sort(key=str.lower)  # Sort files in-place
            for file in files:
                if not file.endswith(image_extension):
                    continue
                print(f"Loading {file}...", end=' ::::::: ')

                img = cv2.imread(os.path.join(root, file))
                assert img is not None, f"Image {file} empty read"
                gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
                # Find the chess board corners
                ret, corners = cv2.findChessboardCorners(gray, (self.cb_width, self.cb_height), None)
                # If found, add object points, image points (after refining them)
                if ret:
                    print(f"Found corners...")
                    objpoints.append(objp)
                    corners2 = cv2.cornerSubPix(gray, corners, (11, 11), (-1, -1), self.criteria)
                    imgpoints.append(corners2)
                    # Draw and display the corners
                    img = cv2.drawChessboardCorners(img, (self.cb_width, self.cb_height), corners2, ret)
                else:
                    print(f"Did not find corners...")
        # flags |= CALIB_ZERO_TANGENT_DIST|CALIB_FIX_K1|CALIB_FIX_K2|CALIB_FIX_K3|CALIB_FIX_K4|CALIB_FIX_K5|CALIB_FIX_K6
        ret, mtx, dist, rvecs, tvecs = cv2.calibrateCamera(objpoints, imgpoints, gray.shape[::-1], None, None)
        h, w, channel = img.shape
        mtx, roi = cv2.getOptimalNewCameraMatrix(mtx, dist, (w, h), 1, (w, h))
        print("Calibration is finished. RMS: ", ret)
        self.camera_matrix = mtx
        self.PRINCIPAL_POINT = np.array([mtx[0, 2], mtx[1, 2]])
        self.FOCAL_LENGTH = mtx[0, 0]
        self.PIXEL_ASPECT_RATIO = mtx[1, 1] / mtx[0, 0]
        self.dist_coeff = dist
        # self.rvecs = rvecs
        # self.tvecs = tvecs
        # self.roi = roi
        self.RESOLUTION = (w, h)
        self.RESOLUTION_U = w
        self.RESOLUTION_V = h
        self.SKEW = mtx[0, 1]
        self.RADIAL_DISTORTION1, self.RADIAL_DISTORTION2, self.RADIAL_DISTORTION3, self.TANGENTAL_DISTORTION1, self.TANGENTAL_DISTORTION2 \
            = dist.reshape(-1)
        return [ret, mtx, dist, rvecs, tvecs]

    def extrinsic_calibrate(self, verbose=True):
        success, rotation_vector, translation_vector = cv2.solvePnP(self.carpet_3D, self.carpet_2D, self.camera_matrix, self.dist_coeff, flags=0)
        print("Extrinsic calibration is finished. Success: ", success)
        self.translation_vector = translation_vector
        self.rotation_vector = rotation_vector
        self.rot3x3, _ = cv2.Rodrigues(rotation_vector)
        R_inv = self.rot3x3.T  # Transpose of rotation matrix is its inverse
        t_world = -R_inv @ translation_vector
        self.POSITION = t_world

    def set_carpet_3D(self, type='checkerboard', quadrant=1):
        """
        Simplified version:
        1) assume checkerboard is a carpet on the ground and orthogonal to gravity direction, z is up
        2) set detected checkerboard corners_2d[0] is world origin
        3) quadrant 1, 2, 3, 4, representing which quadrant the checkerboard is in regards to the +x, +y 1st quadrant in world coord
        """
        # set 3d carpet keypoints
        if type=='checkerboard':
            """ use the same checkerboard (with intrinsic calibration) to link camera to world coord"""
            # define world coord
            raise NotImplementedError

        elif type=='markers':
            """ use tape markers """
            raise NotImplementedError
    def set_checkerboard_carpet_3D(self, cam_pos='back'):
        """
        cam_position: 'back', 'right', 'left'
        Easier to customize this function each time you use
        """
        width = self.cb_width
        height = self.cb_height
        carpet_3D = np.zeros((height, width, 3))
        for h in range(height):
            for w in range(width):
                if cam_pos=='back':
                    carpet_3D[h, w, 0] = -w
                    carpet_3D[h, w, 1] = h
                elif cam_pos=='right':
                    carpet_3D[h, w, 0] = -h
                    carpet_3D[h, w, 1] = -w
                elif cam_pos == 'left':
                    carpet_3D[h, w, 0] = h
                    carpet_3D[h, w, 1] = w
                else:
                    raise ValueError(f"cam_pos {cam_pos} not recognized")
        carpet_3D = carpet_3D.reshape([-1, 3])
        carpet_3D = carpet_3D * self.square_size
        self.carpet_3D = carpet_3D
        return carpet_3D


    def find_carpet_2D(self, img_path, verbose=False):
        img = cv2.imread(img_path)
        assert img is not None, f"Image {file} empty read"

        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        # Find the chess board corners
        ret, corners = cv2.findChessboardCorners(gray, (self.cb_width, self.cb_height), None)

        self.got_carpet_2D = ret
        # If found, add object points, image points (after refining them)
        if ret:
            print(f"Found corners...")
            corners2 = cv2.cornerSubPix(gray, corners, (11, 11), (-1, -1), self.criteria)
            # Draw and display the corners

            if verbose:
                img = cv2.drawChessboardCorners(img, (self.cb_width, self.cb_height), corners2, ret)
                # draw corner[0] as origin
                cv2.circle(img, tuple(corners2[0].astype(int).ravel()), 10, (255, 255, 255), -1)
                cv2.imshow('img', img)
                cv2.waitKey(0)

            h, w, channel = img.shape
            if not w == self.RESOLUTION_U:
                # stupid iphone's camera function in video mode gives bigger image than video frames
                ratio = self.RESOLUTION_U / w
                print(f"Image resolution does not match camera resolution: {img.shape} vs {self.RESOLUTION}, rescaling corners by {ratio}")
                assert self.RESOLUTION_V / h == ratio, f"Image landscape/portrait does not match camera resolution: {img.shape} vs {self.RESOLUTION}, not implemented yet"
                corners2 = corners2 * ratio
            self.carpet_2D = corners2
        else:
            print(f"Did not find corners...")
        return ret



    def save(self, save_dir):
        with open(save_dir, 'wb') as f:
            pickle.dump(self.__dict__, f)

    def load(self, load_dir):
        with open(load_dir, 'rb') as f:
            self.__dict__ = pickle.load(f)

    def undistort(self, points_2d):
        raise NotImplementedError

    def distort(self, points_2d):
        raise NotImplementedError





if __name__ == '__main__':
    example_case = 1
    if example_case == 1:  # Camera intrinsic and extrinsic from checkerboard images
        checkerboard_dir = r'/Volumes/Z/Onform/synthetic_camera/calibrate/1x_intrinsic'
        # set up camera
        camera_base = Checkerboard_Camera()
        camera_base.set_checkerboard_params(square_size=0.0243, width=9, height=6)
        camera_base.intrinsic_calibrate(checkerboard_dir, '.png')

        extrinsic_dir = r'/Volumes/Z/Onform/synthetic_camera/calibrate/1x_extrinsic'
        image_extension = '.JPG'
        device_id = 0
        for root, dirs, files in os.walk(extrinsic_dir):
            files = [f for f in files if not f[0] == '.']  # skip hidden files
            for file in files:
                if not file.endswith(image_extension):
                    continue
                if "frames" in root:
                    continue
                print(f"Loading {file}...", end=' ::::::: ')
                camera_extrinsic = camera_base.copy()
                camera_extrinsic.find_carpet_2D(os.path.join(root, file), verbose=False)

                location_of_interest = ['back', 'right', 'left']
                device_id += 1
                for location in location_of_interest:
                    camera = camera_extrinsic.copy()
                    camera.set_camera_params(DEVICEID=f"{location}_{device_id}", VIEW_ROTATION=0)
                    camera.set_checkerboard_carpet_3D(cam_pos=location)

                    camera.extrinsic_calibrate()
                    camera.save(os.path.join(extrinsic_dir, f"{file}_{device_id}_{location}" + '.pkl'))
                    print(camera.POSITION)
                    proj_point2D, jacobian = cv2.projectPoints(camera.carpet_3D, camera.rotation_vector, camera.POSITION,
                                                               camera.camera_matrix, camera.dist_coeff)
                    # # plot
                    # img = cv2.imread(os.path.join(root, file))
                    # img = cv2.drawChessboardCorners(img, (camera.cb_width, camera.cb_height), proj_point2D, True)
                    # cv2.imshow('img', img)
                    # cv2.waitKey(0)




from .Camera import *

def batch_load_from_theia3d(xml_filename, camera_models=["Miqus Video"]):
    tree = ET.parse(xml_filename)
    root = tree.getroot()
    cameras = []
    for child in root:
        if child.tag == 'cameras':
            for grandchild in child:
                camera = Theia3d_Camera()
                camera.load_theia3d_xml(grandchild)
                if camera.DEVICE_MODEL in camera_models:
                    cameras.append(camera)
    return cameras


class Theia3d_Camera(Camera):
    """
    A class representing a camera in Theia3D format.
    https://www.theiamarkerless.ca/docs/dataformat.html
    """
    def __init__(self):
        super().__init__()

    def load_theia3d_xml(self, xml_element):
        # Read the xcp file as xml
        self.DEVICEID = xml_element.attrib['serial']
        self.DEVICE_MODEL = xml_element.attrib['model']
        self.VIEW_ROTATION = float(xml_element.attrib['viewrotation'])
        if xml_element.attrib['video_resolution'] == '540p':
            self.RESOLUTION = np.array([960, 544])
        elif xml_element.attrib['video_resolution'] == '1080p':
            self.RESOLUTION = np.array([1920, 1080])
        else:
            raise NotImplementedError
        self.RESOLUTION_U = self.RESOLUTION[0]
        self.RESOLUTION_V = self.RESOLUTION[1]

        transform = xml_element.find('transform')
        self.X = float(transform.attrib['x'])
        self.Y = float(transform.attrib['y'])
        self.Z = float(transform.attrib['z'])
        self.R11 = float(transform.attrib['r11'])
        self.R12 = float(transform.attrib['r12'])
        self.R13 = float(transform.attrib['r13'])
        self.R21 = float(transform.attrib['r21'])
        self.R22 = float(transform.attrib['r22'])
        self.R23 = float(transform.attrib['r23'])
        self.R31 = float(transform.attrib['r31'])
        self.R32 = float(transform.attrib['r32'])
        self.R33 = float(transform.attrib['r33'])
        self.rot3x3 = np.array([[self.R11, self.R12, self.R13],
                                [self.R21, self.R22, self.R23],
                                [self.R31, self.R32, self.R33]])
        # self.ORIENTATION = self.quaternion_from_rot3x3(self.rot3x3), # just use rot3x3 for now
        self.POSITION = np.array([self.X, self.Y, self.Z])

        # Extracting the 'intrinsic' attributes
        intrinsic = xml_element.find('intrinsic')
        # self.FOCAL_LENGTH = float(intrinsic.attrib['focallength'])  # not sure what this is, 9.223126, using sensor to px ratio instead
        self.SENSOR_MIN_U = float(intrinsic.attrib['sensorMinU'])
        self.SENSOR_MAX_U = float(intrinsic.attrib['sensorMaxU'])
        self.SENSOR_MIN_V = float(intrinsic.attrib['sensorMinV'])
        self.SENSOR_MAX_V = float(intrinsic.attrib['sensorMaxV'])
        self.FOCAL_LENGTH_U_SENSOR = float(intrinsic.attrib['focalLengthU'])
        self.FOCAL_LENGTH_V_SENSOR = float(intrinsic.attrib['focalLengthV'])
        self.CENTER_POINT_U_SENSOR = float(intrinsic.attrib['centerPointU'])
        self.CENTER_POINT_V_SENSOR = float(intrinsic.attrib['centerPointV'])
        self.SKEW = float(intrinsic.attrib['skew'])  # 0.0
        self.RADIAL_DISTORTION1 = float(intrinsic.attrib['radialDistortion1'])
        self.RADIAL_DISTORTION2 = float(intrinsic.attrib['radialDistortion2'])
        self.RADIAL_DISTORTION3 = float(intrinsic.attrib['radialDistortion3'])
        self.TANGENTAL_DISTORTION1 = float(intrinsic.attrib['tangentalDistortion1'])
        self.TANGENTAL_DISTORTION2 = float(intrinsic.attrib['tangentalDistortion2'])

        sensor_U = self.SENSOR_MAX_U - self.SENSOR_MIN_U
        sensor_V = self.SENSOR_MAX_V - self.SENSOR_MIN_V
        pp_U_ratio = self.CENTER_POINT_U_SENSOR / sensor_U
        pp_V_ratio = self.CENTER_POINT_V_SENSOR / sensor_V
        px_to_sensor_ratio_U = self.RESOLUTION_U / sensor_U
        px_to_sensor_ratio_V = self.RESOLUTION_V / sensor_V
        assert abs(pp_U_ratio - 0.5) < 0.1, f'pp_U_ratio: {pp_U_ratio}'
        assert abs(pp_V_ratio - 0.5) < 0.1, f'pp_V_ratio: {pp_V_ratio}'
        assert abs(px_to_sensor_ratio_U - px_to_sensor_ratio_V) < 0.01, f'px_to_sensor_ratio_U: {px_to_sensor_ratio_U}, px_to_sensor_ratio_V: {px_to_sensor_ratio_V}'
 
        self.PRINCIPAL_POINT = np.array([self.RESOLUTION_U*pp_U_ratio, self.RESOLUTION_V*pp_V_ratio])
        self.DISTORTION_CENTER = self.PRINCIPAL_POINT
        self.DISTORTION_SCALE = np.array([1.0, self.RADIAL_DISTORTION1, self.RADIAL_DISTORTION2, self.RADIAL_DISTORTION3])
        self.PIXEL_ASPECT_RATIO = self.FOCAL_LENGTH_V_SENSOR / self.FOCAL_LENGTH_U_SENSOR

        self.FOCAL_LENGTH_U = self.FOCAL_LENGTH_U_SENSOR * px_to_sensor_ratio_U
        self.FOCAL_LENGTH_V = self.FOCAL_LENGTH_V_SENSOR * px_to_sensor_ratio_V
        self.FOCAL_LENGTH = self.FOCAL_LENGTH_U  # the projection fuction will multiply U with skew factor for V

    def undistort(self, points_2d):  # todo: modify for theia3d
        raise NotImplementedError
        # undistort radial distortion, formula provided by Vicon
        points_2d = np.array(points_2d)
        dp = (points_2d - self.DISTORTION_CENTER) * (np.array([1, self.PIXEL_ASPECT_RATIO]))
        radius = np.linalg.norm(dp, axis=1)
        scale = np.vstack([np.ones_like(radius), radius ** 2, radius ** 4, radius ** 6]).T.dot(self.DISTORTION_SCALE).reshape(-1, 1)
        point_2d_corrected = (scale*dp + self.DISTORTION_CENTER)*(np.array([1, 1/self.PIXEL_ASPECT_RATIO]))  # assuming distortion center is the same as principal point
        return point_2d_corrected

    # def distort(self, points_2d):
    #     # approximate distort radial distortion, scale calculation should use distorted points, but here we use undistorted points
    #     points_2d = np.array(points_2d)
    #     dp = points_2d * (np.array([1, self.PIXEL_ASPECT_RATIO])) - self.DISTORTION_CENTER
    #     radius = np.linalg.norm(dp, axis=1)
    #     scale = np.vstack([np.ones_like(radius), radius ** 2, radius ** 4, radius ** 6]).T.dot(self.DISTORTION_SCALE).reshape(-1, 1)
    #     point_2d_distorted = dp/scale*(np.array([1, 1/self.PIXEL_ASPECT_RATIO])) + self.DISTORTION_CENTER # assuming distortion center is the same as principal point
    #     return point_2d_distorted

    def distort(self, points_2d):
        raise NotImplementedError
        # Convert points to numpy array
        points_2d = np.array(points_2d)

        # Extract X and Y coordinates
        x = points_2d[:, 0]
        y = points_2d[:, 1]

        # Normalize points by the principal point (assuming principal point is the distortion center)
        x = (x - self.DISTORTION_CENTER[0])
        y = (y - self.DISTORTION_CENTER[1])

        # Compute radius squared
        r_squared = x ** 2 + y ** 2

        # Radial distortion correction
        radial_distortion = (1 + self.RADIAL_DISTORTION1 * r_squared +
                             self.RADIAL_DISTORTION2 * r_squared ** 2 +
                             self.RADIAL_DISTORTION3 * r_squared ** 3)

        x_radial = x #* radial_distortion
        y_radial = y #* radial_distortion

        r_squared = x_radial ** 2 + y_radial ** 2
        # Tangential distortion correction
        x_tangential = 2 * self.TANGENTAL_DISTORTION1 * x_radial * y_radial + self.TANGENTAL_DISTORTION2 * (r_squared + 2 * x_radial ** 2)
        y_tangential = self.TANGENTAL_DISTORTION1 * (r_squared + 2 * y_radial ** 2) + 2 * self.TANGENTAL_DISTORTION2 * x_radial * y_radial

        # Apply corrections
        x_corrected = x_radial + x_tangential
        y_corrected = y_radial + y_tangential

        # Re-center the points back to original coordinates
        x_corrected += self.DISTORTION_CENTER[0]
        y_corrected += self.DISTORTION_CENTER[1]

        # Return the distorted points
        return np.column_stack((x_corrected, y_corrected))




if __name__ == '__main__':
    example_case = 1
    if example_case == 1:
        xml_filename = r'H:\Onform\render\Destrie Grieman - 1\Fastball RH Markerless 1.settings.xml'
        cameras = batch_load_from_theia3d(xml_filename)
        
